
;;; Destructuring package:  Subsumes let, let* and set!.
;;; written by Gary Brooks
;;; 6/18/84

;;; Allows for 2lisp-like destructuring.
;;;
;;;
;;; Desctructuring patterns can be of the form:
;;;
;;;    <pattern> ::=  <atom> | <pattern> X .... X <pattern>
;;; 
;;; That is a pattern is either atomic or a crossproduct of patterns.
;;; &rest or dot, ".", destructuring, found in various dialects of Lisp, 
;;; are not supported.  Dot destructuring operates on elements of D*.
;;; The destructuring supported by this package is viewed as operating only
;;; on elements of D^n.
;;;
;;; The destructuring macros:
;;;
;;; (seq item ...)   
;;;
;;; Constructs a crossproduct of items, i.e. an element of D^n.
;;;
;;; 
;;; (mvlet
;;;    [[<pattern1> <value1>] 
;;;	...
;;;	[<patternN> <valueN>]]
;;;    <body>)
;;;
;;;  Binds the atomic elements (variables) of each pattern to the 
;;;  corresponding elements of each value.  Like let, all the values
;;;  are evaluated in an environment free of the pattern variables.
;;; 
;;; (mvlet*
;;;    [[<pattern1> <value1>] 
;;;	...
;;;	[<patternN> <valueN>]]
;;;    <body>)
;;;
;;;  Like mvlet, except that subsequent values are evaluated in an 
;;;  environment augmented by previous pattern variables.
;;;
;;;  Examples:
;;; 
;;; >>> (mvlet 
;;;        [[b (seq (seq 'a 'b) (seq 'c 'd))]]  ;;; '[[a b] [c d]]
;;;        (mvlet 
;;; 	  [[[b [y z]] b]   ;;; this binding to b doesn't
;;; 	   [[i j] b]]      ;;; effect this use of b.
;;; 	  (seq b i j y z)))
;;; ((a b) (a b) (c d) c d)
;;; >>> (mvlet
;;;        [[b (seq (seq 'a 'b) (seq 'c 'd))]]  ;;; '[[a b] [c d]]
;;;        (mvlet*
;;; 	  [[[b [y z]] b]   ;;; b is being rebound
;;; 	   [[i j] b]]      ;;; and this b uses the new value.
;;; 	  (seq b i j y z)))
;;; ((a b) a b c d)
;;; 


(macro seq (lambda (form) `(list ,@(cdr form))))

(letrec
   ((map-vars 
       (lambda (vars body)
	  (if (null? vars) 
	      (seq nil body)
	      (let [[v (map-vars (cdr vars) body)]]
		 (let [[vvars (car v)]
		       [vbody (cadr v)]]
		    (let [[a (make-var (car vars) vbody)]]
		       (let [[avar (car a)]
			     [abody (cadr a)]]
			  (seq (cons avar vvars)
			       abody))))))))
    (make-var
       (lambda (vars body)
	  (if (atom? vars)
	      (seq vars body)
	      (let [[p (map-vars vars body)]]
		 (let [[pvars (car p)]
		       [pbody (cadr p)]]
		    (let [[tvar (gensym 'm)]]
		       (seq tvar 
			  `(apply (lambda ,pvars
				     ,pbody)
				  ,tvar)))))))))
   (letrec 
      ((make-let 	  
	  (lambda (var-lists body)
	     (if (null? var-lists)
		 (seq nil nil body)
		 (let ((ll (car var-lists)))
		    (if (or (atom? ll)
			    (not (= (length ll) 2)))
			(writeln "Bad let pair " ll))
		    (let ((vars (car ll))
			  (exp (cadr ll)))
		       (let [[x (make-var vars body)]]
			  (let [[xvar (car x)]
				[xbody (cadr x)]]
			     (let [[y (make-let (cdr var-lists) xbody)]]
				(let [[yvars (car y)]
				      [yvals (cadr y)]
				      [ybody (caddr y)]]
				   (seq (cons xvar yvars)
					(cons exp yvals)
					ybody)))))))))))
      (macro mvlet
	 (lambda (form)
	    (let [[x (make-let (cadr form) (cons 'begin (cddr form)))]]
	       `((lambda ,(car x) ,(caddr x))
		 .,(cadr x))))))   
   
   
   (letrec 
      ((make-let* 	  
	  (lambda (var-lists body)
	     (if (null? var-lists)
		 body
		 (let ((ll (car var-lists)))
		    (if (or (atom? ll)
			    (not (= (length ll) 2)))
			(writeln "Bad let pair " ll))
		    (let ((vars (car ll))
			  (exp (cadr ll)))
		       (let [[y (make-let* (cdr var-lists) body)]]
			  (let [[x (make-var vars y)]]
			     (let [[xvar (car x)]
				   [xbody (cadr x)]]
				`((lambda (,xvar) 
				     ,xbody)
				  ,exp))))))))))
      (macro mvlet*
	 (lambda (form)
	    (make-let* (cadr form) (cons 'begin (cddr form)))))))


(letrec
   ((map-vars 
       (lambda (vars)
	  (if (null? vars)
	      (seq nil nil)
	      (let [[v (map-vars (cdr vars))]]
		 (let [[vvars (car v)]
		       [vbody (cadr v)]]
		    (let [[tvar (gensym 'm)]]
			(let [[abody (make-var (car vars) tvar)]]
			      (seq (cons tvar vvars)
				   (cons abody vbody)))))))))
    (make-var
       (lambda (vars body)
	  (if (atom? vars)
	      `(set! ,vars ,body)
	      (let [[v (map-vars vars)]]
		 `(apply (lambda ,(car v)
			    (begin .,(cadr v)))
			 ,body))))))
   (macro mvset! 
      (lambda (form)
	 (make-var (cadr form) (caddr form)))))
			
			 
